#include "MeshLoader.h"

const std::vector<Node>& MeshLoader::getNodeSTL() const 
{
    return nodeSTL;
}

const std::vector<FiniteElement>& MeshLoader::getFiniteSTL() const 
{
    return finiteSTL;
}

const std::vector<BoundaryFiniteElement>& MeshLoader::getBoundarySTL() const 
{
    return boundarySTL;
}

std::vector<size_t> MeshLoader::findFiniteElementById(int x, int y, int z) const 
{
    std::vector<size_t> element{};
    auto pred = [x, y, z](const FiniteElement& FE)
    {
        return std::find(FE.FiniteElementIdList.begin(), FE.FiniteElementIdList.end(), x) != FE.FiniteElementIdList.end()
            && std::find(FE.FiniteElementIdList.begin(), FE.FiniteElementIdList.end(), y) != FE.FiniteElementIdList.end()
            && std::find(FE.FiniteElementIdList.begin(), FE.FiniteElementIdList.end(), z) != FE.FiniteElementIdList.end();
    };
    auto iter = finiteSTL.begin();
    while (iter != finiteSTL.end()) 
    {
        iter = std::find_if(iter, finiteSTL.end(), pred);
        if (iter != finiteSTL.end()) 
        {
            element.push_back(iter->id);
            ++iter;
        }
    }
    return element;
}

std::vector<FiniteElement> MeshLoader::findFiniteElementByEdge(int x, int y) const 
{
    std::vector<FiniteElement> element{};
    auto pred = [x, y](const FiniteElement& FE)
    {
        return std::find(FE.FiniteElementIdList.begin(), FE.FiniteElementIdList.end(), x) != FE.FiniteElementIdList.end()
            && std::find(FE.FiniteElementIdList.begin(), FE.FiniteElementIdList.end(), y) != FE.FiniteElementIdList.end();
    };
    auto iter = finiteSTL.begin();
    while (iter != finiteSTL.end()) 
    {
        iter = std::find_if(iter, finiteSTL.end(), pred);
        if (iter != finiteSTL.end()) 
        {
            element.push_back(*iter);
            ++iter;
        }
    }
    return element;
}

std::vector<BoundaryFiniteElement> MeshLoader::getBoundaryFiniteElementSTLByBoundaryId(size_t boundaryId) const 
{
    std::vector<BoundaryFiniteElement> element{};
    auto pred = [boundaryId](const BoundaryFiniteElement& BFE) 
    {
        return BFE.boundaryId == boundaryId;
    };
    auto iter = boundarySTL.begin();
    while (iter != boundarySTL.end()) 
    {
        iter = std::find_if(iter, boundarySTL.end(), pred);
        if (iter != boundarySTL.end()) 
        {
            element.push_back(*iter);
            ++iter;
        }
    }
    return element;
}

std::set<size_t> MeshLoader::getNodeSTLByBoundaryId(size_t boundaryId) const 
{
    std::set<size_t> element{};
    auto pred = [boundaryId](const BoundaryFiniteElement& BFE) 
    {
        return BFE.boundaryId == boundaryId;
    };
    auto iter = boundarySTL.begin();
    while (iter != boundarySTL.end()) 
    {
        iter = std::find_if(iter, boundarySTL.end(), pred);
        if (iter != boundarySTL.end()) 
        {
            for (const auto& node : iter->BoundaryFiniteElementIdList) 
                element.insert(node);
            ++iter;
        }
    }
    return element;
}

std::vector<FiniteElement> MeshLoader::getFiniteElementByMaterial(size_t MaterialId) const 
{
    std::vector<FiniteElement> element{};
    auto pred = [MaterialId](const FiniteElement& FE) 
    {
        return MaterialId == FE.MaterialId;
    };
    auto iter = finiteSTL.begin();
    while (iter != finiteSTL.end()) 
    {
        iter = std::find_if(iter, finiteSTL.end(), pred);
        if (iter != finiteSTL.end()) 
        {
            element.push_back(*iter);
            ++iter;
        }
    }
    return element;
}

std::vector<std::set<size_t>> MeshLoader::getNeighbourSTL() const 
{
    std::vector<std::set<size_t>> vector(nodeSTL.size() + 1);
    for (const auto& element : finiteSTL)
        for (const auto& id : element.FiniteElementIdList)
            for (const auto& neighbourId : element.FiniteElementIdList)
                if (id != neighbourId)
                    vector[id].insert(neighbourId);
    return vector;
}

void MeshLoader::insertNodesInCenter()
{
    std::unordered_set <Edge, Hash> edge_set;
    for (auto& element : finiteSTL) 
    {
        std::vector<size_t> FiniteElementNodeId = element.FiniteElementIdList;
        for (int first = 0; first < 3; ++first) 
        {
            for (auto second = first + 1; second < 4; ++second) 
            {
                Edge currentEdge(FiniteElementNodeId[first], FiniteElementNodeId[second]);
                if (edge_set.insert(currentEdge).second) // check if such point already exists
                {
                    Node new_Node = getCenterNode(currentEdge);
                    currentEdge.setCenter(new_Node.id);
                    nodeSTL.push_back(new_Node);
                    element.FiniteElementIdList.push_back(new_Node.id);
                }
                else 
                {
                    element.FiniteElementIdList.push_back(currentEdge.centerNode);
                }
            }
        }
    }
    for (auto& element : finiteSTL) 
    {
        std::vector<size_t> BoundaryFiniteElementNodeId = element.FiniteElementIdList;
        for (auto first = 0; first < 3; ++first) 
        {
            for (auto second = first + 1; second < 3; ++second) 
            {
                auto currentEdgeIter = edge_set.find({BoundaryFiniteElementNodeId[first], BoundaryFiniteElementNodeId[second]});
                if (currentEdgeIter == edge_set.end()) 
                {
                    currentEdgeIter = edge_set.find({BoundaryFiniteElementNodeId[second], BoundaryFiniteElementNodeId[first]});
                }
                element.FiniteElementIdList.push_back(currentEdgeIter->centerNode);
            }
        }
    }
}

std::array<double, 3> MeshLoader::getCenterEdge(const Edge& edge) 
{
    double centerX = ((nodeSTL.at(edge.nodePair.first - 1).coordinates.at(0)) +
        (nodeSTL.at(edge.nodePair.second - 1).coordinates.at(0))) / 2;
    double centerY = ((nodeSTL.at(edge.nodePair.first - 1).coordinates.at(1)) +
        (nodeSTL.at(edge.nodePair.second - 1).coordinates.at(1))) / 2;
    double centerZ = ((nodeSTL.at(edge.nodePair.first - 1).coordinates.at(2)) +
        (nodeSTL.at(edge.nodePair.second - 1).coordinates.at(2))) / 2;
    return {centerX, centerY, centerZ};
}

Node MeshLoader::getCenterNode(const Edge& edge) 
{
    Node new_Node(nodeSTL.size() + 1, getCenterEdge(edge), false); // using new Node constructor
    return new_Node;
}